from pulr import config, register_puller, set_data

from pulr.dp import (parse_int, prepare_transform, run_transform, DATA_TYPE_BIT,
                     DATA_TYPE_INT16, DATA_TYPE_INT32, DATA_TYPE_REAL32,
                     DATA_TYPE_UINT16, DATA_TYPE_UINT32, DATA_TYPE_UINT64,
                     MAX_INT16, MAX_INT32)

from functools import partial
import struct

import pymodbus.client.sync
import jsonschema

client = None

SCHEMA_PROTO = {
    'type': 'object',
    'properties': {
        'name': {
            'type': 'string',
            'enum': ['modbus/tcp', 'modbus/udp']
        },
        'source': {
            'type': 'string'
        },
        'default-unit': {
            'type': ['integer', 'string']
        }
    },
    'additionalProperties': False,
    'required': ['name', 'source']
}

SCHEMA_PULL = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'reg': {
                'type': 'string'
            },
            'count': {
                'type': 'integer',
                'minimum': 1
            },
            'unit': {
                'type': ['integer', 'string']
            },
            'process': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'offset': {
                            'type': ['integer', 'string'],
                        },
                        'set-id': {
                            'type': 'string',
                        },
                        'type': {
                            'type':
                                'string',
                            'enum': [
                                'real', 'real32', 'uint16', 'word', 'uint32',
                                'dword', 'sint16', 'int16', 'sint32', 'int32'
                            ]
                        },
                        'transform': {
                            'type': 'array'
                        }
                    },
                    'additionalProperties': False,
                    'required': ['offset', 'set-id']
                }
            }
        },
        'additionalProperties': False,
        'required': ['reg', 'count', 'process']
    }
}

# data conversion functions


def value_to_data(o, offset, transform, data_in):
    value = data_in[offset]
    if transform is not None:
        value = run_transform(transform, None, value)
    set_data(o, value)


def int16_to_data(o, offset, signed, transform, data_in):
    value = data_in[offset]
    if signed and value > MAX_INT16:
        value -= 65536
    if transform is not None:
        value = run_transform(transform,
                              DATA_TYPE_INT16 if signed else DATA_TYPE_UINT16,
                              value)
    set_data(o, value)


def int32_to_data(o, offset, signed, transform, data_in):
    value = data_in[offset] * 65536 + data_in[offset + 1]
    if signed and value > MAX_INT32:
        value -= 4294967296
    if transform is not None:
        value = run_transform(transform,
                              DATA_TYPE_INT32 if signed else DATA_TYPE_UINT32,
                              value)
    set_data(o, value)


def real32_to_data(o, offset, transform, data_in):
    value = struct.unpack(
        'f',
        struct.pack('H', data_in[offset]) +
        struct.pack('H', data_in[offset + 1]))[0]
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_REAL32, value)
    set_data(o, value)


def bit_to_data(o, offset, bit, transform, data_in):
    value = (data_in[offset] >> bit) & 1 == 1
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_BIT, value)
    set_data(o, value)


def parse_offset(offset, addr):
    if isinstance(offset, int):
        bit = None
    else:
        if '/' in offset:
            offset, bit = offset.split('/')
            bit = int(bit)
        else:
            bit = None
        if offset.startswith('='):
            absolute = True
            offset = offset[1:]
        else:
            absolute = False
        offset = parse_int(offset)
        if absolute:
            offset -= addr
    if offset < 0:
        raise ValueError('offset is negative')
    if bit is not None and bit < 0:
        raise ValueError('bit is negative')
    return offset, bit


def process_data(fn, dtype):
    rr = fn()
    if rr.isError():
        raise Exception(f'Modbus error {fn}')
    return rr.bits if dtype == 'b' else rr.registers


def init(cfg_proto, cfg_pull, timeout=5):
    global client
    jsonschema.validate(cfg_proto, SCHEMA_PROTO)
    jsonschema.validate(cfg_pull, SCHEMA_PULL)
    if cfg_proto['name'] in ['modbus/tcp', 'modbus/udp']:
        try:
            host, port = cfg_proto['source'].rsplit(':', 1)
        except:
            host = cfg_proto['source']
            port = 502
        if cfg_proto['name'] == 'modbus/tcp':
            client = pymodbus.client.sync.ModbusTcpClient(host, int(port))
        else:
            client = pymodbus.client.sync.ModbusUdpClient(host, int(port))
    else:
        raise ValueError(f'Unsupported protocol: {cfg_proto["name"]}')

    client.timeout = timeout

    for p in cfg_pull:
        try:
            unit = p['unit']
        except KeyError:
            unit = cfg_proto.get('default-unit', 1)
        u = parse_int(unit)
        reg = p['reg']
        addr = parse_int(reg[1:])
        if reg[0] == 'c':
            pfn = client.read_coils
        elif reg[0] == 'd':
            pfn = client.read_discrete_inputs
        elif reg[0] == 'h':
            pfn = client.read_holding_registers
        elif reg[0] == 'i':
            pfn = client.read_input_registers
        else:
            raise ValueError(f'Invalid register type: {reg[0]}')
        pmap = []
        for m in p.get('process', []):
            offset = m['offset']
            o = m['set-id']
            tp = m.get('type')
            transform = m.get('transform')
            if reg[0] in ['h', 'i']:
                offset, bit = parse_offset(offset, addr)
                if bit is None:
                    if tp in ['real', 'real32']:
                        fn = partial(real32_to_data, o, offset,
                                     prepare_transform(o, transform))
                    elif not tp or tp in ['uint16', 'word']:
                        fn = partial(int16_to_data, o, offset, False,
                                     prepare_transform(o, transform))
                    elif tp in ['uint32', 'dword']:
                        fn = partial(int32_to_data, o, offset, False,
                                     prepare_transform(o, transform))
                    elif tp in ['sint16', 'int16']:
                        fn = partial(int16_to_data, o, offset, True,
                                     prepare_transform(o, transform))
                    elif tp in ['sint32', 'int32']:
                        fn = partial(int32_to_data, o, offset, True,
                                     prepare_transform(o, transform))
                    else:
                        raise ValueError(f'type unsupported: {tp}')
                else:
                    fn = partial(bit_to_data, o, offset, bit,
                                 prepare_transform(o, transform))
            else:
                fn = partial(value_to_data, o, offset,
                             prepare_transform(o, transform))
            pmap.append(fn)
        register_puller(
            partial(process_data,
                    partial(pfn, addr, count=p.get('count', 1), unit=u),
                    ('b' if reg[0] in ['c', 'd'] else 'w')), pmap)
    client.connect()


def shutdown():
    client.close()



from pulr import set_data, get_last_pull_time

from functools import partial


def parse_int(i):
    if isinstance(i, int):
        return i
    elif isinstance(i, float):
        return int(i)
    else:
        if '+' in i:
            r = 0
            for x in i.split('+'):
                r += parse_int(x)
            return r
        elif 'x' in i:
            return int(i, 16)
        else:
            return int(i)


# common data types

MAX_INT16 = 32767
MAX_INT32 = 2147483647

MAX_UINT8 = 255
MAX_UINT16 = 65535
MAX_UINT32 = 4294967295
MAX_UINT64 = 18446744073709551615

DATA_TYPE_BIT = 0

DATA_TYPE_INT8 = 1
DATA_TYPE_INT16 = 2
DATA_TYPE_INT32 = 3

DATA_TYPE_UINT8 = 10
DATA_TYPE_UINT16 = 11
DATA_TYPE_UINT32 = 12
DATA_TYPE_UINT64 = 13

DATA_TYPE_REAL32 = 20
DATA_TYPE_REAL64 = 21

MAX_UVAL = {
    DATA_TYPE_UINT8: MAX_UINT8,
    DATA_TYPE_UINT16: MAX_UINT16,
    DATA_TYPE_UINT32: MAX_UINT32,
    DATA_TYPE_UINT64: MAX_UINT64
}

# transformers

_speed_cache = {}


def clear():
    _speed_cache.clear()


def transform_speed(o, interval, tp, value):
    maxval = MAX_UVAL[tp]
    if o in _speed_cache:
        v_prev, ptime = _speed_cache[o]
        t_delta = get_last_pull_time() - ptime
        if t_delta < interval:
            return None
        if value >= v_prev:
            v_delta = value - v_prev
        else:
            v_delta = maxval - v_prev + value
        speed = v_delta / t_delta
    else:
        speed = 0
    _speed_cache[o] = (value, get_last_pull_time())
    return speed


def transform_multiply(m, tp, value):
    return value * m


def transform_divide(d, tp, value):
    return value / d


def transform_round(d, tp, value):
    return round(value, None if d == 0 else d)


def transform_bit_to_int(tp, value):
    return 1 if value else 0


def transform_int_to_bit(tp, value):
    return value != 0


def prepare_transform(o, transform):
    if transform is not None:
        transforms = []
        for c in transform:
            if c['type'] == 'speed':
                transforms.append(
                    partial(transform_speed, o, c.get('interval', 1)))
            elif c['type'] == 'multiply':
                transforms.append(partial(transform_multiply, c['multiplier']))
            elif c['type'] == 'divide':
                transforms.append(partial(transform_divide, c['divisor']))
            elif c['type'] == 'round':
                transforms.append(partial(transform_round, c['digits']))
            elif c['type'] == 'bit2int':
                transforms.append(transform_bit_to_int)
            elif c['type'] == 'int2bit':
                transforms.append(transform_int_to_bit)
            else:
                raise ValueError(f'Unsupported transform {c["type"]}')
        return transforms
    else:
        return None


def run_transform(transform, tp, value):
    for c in transform:
        value = c(tp, value)
        if value is None:
            return None
    return value



import sys
import time

from functools import partial
from types import SimpleNamespace

import pytz

from .beacons import beacon_empty_line

oprint = partial(print, flush=True)
eprint = partial(print, flush=True, file=sys.stderr)

time_format = None

try:
    LOCAL_TZ = pytz.timezone(time.tzname[0])
except:
    eprint('unable to determine local time zone')
    LOCAL_TZ = None

TIME_FORMAT_ISO = 1
TIME_FORMAT_TIMESTAMP = 2

TIME_FORMATS = {'iso': TIME_FORMAT_ISO, 'timestamp': TIME_FORMAT_TIMESTAMP}


def set_time_format(tf):
    global time_format
    if tf is not None:
        time_format = TIME_FORMATS[tf]


try:
    import rapidjson as json
except:
    import json


def get_time():
    if time_format is None:
        return None
    elif time_format == TIME_FORMAT_ISO:
        from datetime import datetime
        return datetime.now().replace(tzinfo=LOCAL_TZ).isoformat()
    elif time_format == TIME_FORMAT_TIMESTAMP:
        return time.time()


def print_trace():
    import traceback
    eprint(traceback.format_exc())


def output_devnull(*args, **kwargs):
    pass


def output_stdout(o, value):
    oprint(f'{str(get_time()) + " " if time_format else ""}{o} {value}')


def output_stdout_csv(o, value):
    oprint(f'{str(get_time()) + ";" if time_format else ""}{o};{value}')

def output_stdout_ndjson(o, value):
    d = {'id': o, 'value': value}
    if time_format:
        d['time'] = get_time()
    oprint(json.dumps(d))


def output_eva_datapuller(o, value):
    if o.endswith('.value'):
        s = f'{o[:-6]} u '
        val_mode = True
    elif o.endswith('.status'):
        s = f'{o[:-7]} u '
        val_mode = False
    else:
        s = f'{o} u '
        val_mode = True
    oprint(s + (f'None {value}' if val_mode else str(value)))


OUTPUT_METHODS = {
    None: {
        'output': output_stdout,
        'beacon': beacon_empty_line
    },
    'ndjson': {
        'output': output_stdout_ndjson,
        'beacon': beacon_empty_line
    },
    'csv': {
        'output': output_stdout_csv,
        'beacon': beacon_empty_line
    },
    'eva/datapuller': {
        'output': output_eva_datapuller,
        'beacon': beacon_empty_line
    }
}



from pulr import config, register_puller, set_data
from pulr.dp import (prepare_transform,
                     run_transform, DATA_TYPE_INT32, DATA_TYPE_UINT32,
                     DATA_TYPE_UINT64)
import netsnmp

from functools import partial

import jsonschema

session = None

SCHEMA_PROTO = {
    'type': 'object',
    'properties': {
        'name': {
            'type': 'string',
            'enum': ['snmp']
        },
        'source': {
            'type': 'string'
        },
        'community': {
            'type': ['integer', 'string']
        },
        'version': {
            'type': 'integer',
            'minimal': 1
        },
        'retries': {
            'type': 'integer',
            'minimal': 0
        }
    },
    'additionalProperties': False,
    'required': ['name', 'source']
}

SCHEMA_PULL = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'oids': {
                'type': 'array',
                'items': {
                    'type': 'string'
                }
            },
            'process': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'offset': {
                            'type': ['integer', 'string'],
                        },
                        'oid': {
                            'type': 'string',
                        },
                        'set-id': {
                            'type': 'string',
                        },
                        'transform': {
                            'type': 'array'
                        }
                    },
                    'additionalProperties': False,
                    'required': ['oid']
                }
            },
            'ignore': {
                'type': 'array',
                'items': {
                    'type': 'string'
                }
            }
        },
        'additionalProperties': False,
        'required': ['oids']
    }
}

SNMP_DT = {
    'INTEGER': DATA_TYPE_INT32,
    'INTEGER32': DATA_TYPE_INT32,
    'UNSIGNED32': DATA_TYPE_UINT32,
    'COUNTER': DATA_TYPE_UINT32,
    'COUNTER32': DATA_TYPE_UINT32,
    'COUNTER64': DATA_TYPE_UINT64,
    'GAUGE': DATA_TYPE_UINT32,
    'GAUGE32': DATA_TYPE_UINT32
}

def parse_value(val):
    try:
        value = val.decode()
        try:
            value = int(value)
            if int(value) == float(value):
                value = int(value)
        except:
            pass
    except:
        value = '0x' + ''.join(x[2:].upper() for x in map(hex, val))
    return value

def process_varlist(oid_map, ignore_list, data_in):
    for v in data_in:
        if v.iid != '' and v.iid is not None:
            oid = f'{v.tag}.{v.iid}'
        else:
            oid = v.tag
        if oid in ignore_list:
            continue
        value = parse_value(v.val)
        m = oid_map.get(oid)
        if m:
            if m[0]:
                oid = m[0]
            if m[1] is not None:
                value = run_transform(m[1], SNMP_DT.get(v.type), value)
        set_data(oid, value)


def snmp_get(walk_oids, get_oids):
    result = []
    if walk_oids:
        varlist = netsnmp.VarList(*walk_oids)
        data = session.walk(varlist)
        result += varlist
    if get_oids:
        if len(get_oids) > 1 and session.Version == 2:
            varlist = netsnmp.VarList(*get_oids)
            data = session.getbulk(0, 1, varlist)
            result += varlist
        else:
            for g in get_oids:
                varlist = netsnmp.VarList(g)
                data = session.get(varlist)
                result += varlist
    return result


def init(cfg_proto, cfg_pull, timeout=5):
    global session
    jsonschema.validate(cfg_proto, SCHEMA_PROTO)
    jsonschema.validate(cfg_pull, SCHEMA_PULL)

    try:
        host, port = cfg_proto['source'].rsplit(':', 1)
    except:
        host = cfg_proto['source']
        port = 161

    def get_multiplier(m):
        multiplier = m.get('multiplier')
        if 'divisor' in m:
            if multiplier is not None:
                raise ValueError('both divisor and multiplier specified')
            multiplier = 1 / float(m['divisor'])
        return multiplier

    session = netsnmp.Session(Version=cfg_proto.get('version', 2),
                              DestHost=host,
                              RemotePort=port,
                              Community=cfg_proto.get('community', 'public'),
                              Timeout=int(timeout * 1000000),
                              Retries=cfg_proto.get('retries', 1))
    for p in cfg_pull:
        pfn = partial(snmp_get, [v[:-2] for v in p['oids'] if v.endswith('.*')],
                      [v for v in p['oids'] if not v.endswith('.*')])
        pmap = [
            partial(
                process_varlist, {
                    v['oid']: (v.get('set-id'),
                               prepare_transform(v.get('set-id', v['oid']),
                                                 v.get('transform')))
                    for v in p.get('process', [])
                }, p.get('ignore', []))
        ]
        register_puller(pfn, pmap)


def shutdown():
    pass



# TODO redistribute-compile libplctag in some way

from pulr import config, register_puller, set_data

from pulr.dp import (parse_int, prepare_transform, run_transform, DATA_TYPE_BIT,
                     DATA_TYPE_INT16, DATA_TYPE_INT32, DATA_TYPE_REAL32,
                     DATA_TYPE_UINT16, DATA_TYPE_UINT32, DATA_TYPE_UINT64,
                     DATA_TYPE_INT8, DATA_TYPE_UINT8, DATA_TYPE_REAL64)

from functools import partial
from time import sleep
import ctypes
import platform
import threading

import jsonschema

plc_timeout = 2000
plc_path = ''

system = platform.system()
if system == 'Windows':
    libfile = 'plctag.dll'
elif system == 'Darwin':
    libfile = 'libplctag.dylib'
else:
    libfile = 'libplctag.so'
plc_lib = ctypes.cdll.LoadLibrary(libfile)

plc_tag_create = plc_lib.plc_tag_create
plc_tag_create.restype = ctypes.c_int
plc_tag_create.argtypes = [ctypes.c_char_p, ctypes.c_int]

plc_tag_status = plc_lib.plc_tag_status
plc_tag_status.restype = ctypes.c_int
plc_tag_status.argtypes = [ctypes.c_int]

plc_tag_destroy = plc_lib.plc_tag_destroy
plc_tag_destroy.restype = ctypes.c_int
plc_tag_destroy.argtypes = [ctypes.c_int]

plc_tag_read = plc_lib.plc_tag_read
plc_tag_read.restype = ctypes.c_int
plc_tag_read.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_float32 = plc_lib.plc_tag_get_float32
plc_tag_get_float32.restype = ctypes.c_float
plc_tag_get_float32.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_float64 = plc_lib.plc_tag_get_float64
plc_tag_get_float64.restype = ctypes.c_double
plc_tag_get_float64.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_uint8 = plc_lib.plc_tag_get_uint8
plc_tag_get_uint8.restype = ctypes.c_ubyte
plc_tag_get_uint8.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_int8 = plc_lib.plc_tag_get_int8
plc_tag_get_int8.restype = ctypes.c_byte
plc_tag_get_int8.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_uint16 = plc_lib.plc_tag_get_uint16
plc_tag_get_uint16.restype = ctypes.c_ushort
plc_tag_get_uint16.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_int16 = plc_lib.plc_tag_get_int16
plc_tag_get_int16.restype = ctypes.c_short
plc_tag_get_int16.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_uint32 = plc_lib.plc_tag_get_uint32
plc_tag_get_uint32.restype = ctypes.c_uint
plc_tag_get_uint32.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_int32 = plc_lib.plc_tag_get_int32
plc_tag_get_int32.restype = ctypes.c_int
plc_tag_get_int32.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_uint64 = plc_lib.plc_tag_get_uint64
plc_tag_get_uint64.restype = ctypes.c_uint64
plc_tag_get_uint64.argtypes = [ctypes.c_int, ctypes.c_int]

plc_tag_get_int64 = plc_lib.plc_tag_get_int64
plc_tag_get_int64.restype = ctypes.c_int64
plc_tag_get_int64.argtypes = [ctypes.c_int, ctypes.c_int]

active_tags = {}

SLEEP_STEP = 0.01

SCHEMA_PROTO = {
    'type': 'object',
    'properties': {
        'name': {
            'type': 'string',
            'enum': ['enip/ab_eip']
        },
        'source': {
            'type': 'string'
        },
        'path': {
            'type': 'string'
        },
        'cpu': {
            'type': 'string',
            'enum': ['LGX', 'MLGX', 'PLC', 'MLGX800']
        }
    },
    'additionalProperties': False,
    'required': ['name', 'source', 'cpu']
}

SCHEMA_PULL = {
    'type': 'array',
    'items': {
        'type': 'object',
        'properties': {
            'tag': {
                'type': 'string'
            },
            'size': {
                'type': 'integer',
                'minimal': 1
            },
            'count': {
                'type': 'integer',
                'minimal': 1
            },
            'process': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'offset': {
                            'type': ['integer', 'string'],
                        },
                        'set-id': {
                            'type': 'string',
                        },
                        'type': {
                            'type':
                                'string',
                            'enum': [
                                'real', 'real32', 'uint16', 'word', 'uint32',
                                'dword', 'sint16', 'int16', 'sint32', 'int32',
                                'uint8', 'int8', 'sint8', 'byte', 'int64',
                                'sint64', 'uint64', 'qword', 'real64'
                            ]
                        },
                        'transform': {
                            'type': 'array'
                        }
                    },
                    'additionalProperties': False,
                    'required': ['offset', 'type', 'set-id']
                }
            }
        },
        'additionalProperties': False,
        'required': ['tag', 'process']
    }
}

TAG_STATUS_OK = 0
TAG_STATUS_PENDING = 1


def read_tag(tag_path):
    if tag_path in active_tags:
        tag_id = active_tags[tag_path]
    else:
        tag_id = plc_tag_create(ctypes.c_char_p((plc_path + tag_path).encode()),
                                plc_timeout)
        if tag_id < 0:
            raise RuntimeError(f'{tag_path} error {tag_id}')
        while True:
            rc = plc_tag_status(tag_id)
            if rc == TAG_STATUS_PENDING:
                sleep(SLEEP_STEP)
                continue
            elif rc != TAG_STATUS_OK:
                raise RuntimeError(f'{tag_path} status error {rc}')
            break
        active_tags[tag_path] = tag_id
    rc = plc_tag_read(active_tags[tag_path], plc_timeout)
    if rc != TAG_STATUS_OK:
        raise RuntimeError(tag_path)
    return tag_id


def real32_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_float32(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_REAL32, value)
    set_data(o, value)


def real64_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_float64(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_REAL32, value)
    set_data(o, value)


def uint8_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_uint8(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_UINT8, value)
    set_data(o, value)


def int8_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_int8(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_INT8, value)
    set_data(o, value)


def uint16_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_uint16(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_UINT16, value)
    set_data(o, value)


def int16_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_int16(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_INT16, value)
    set_data(o, value)


def uint32_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_uint32(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_UINT32, value)
    set_data(o, value)


def int32_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_int32(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_INT32, value)
    set_data(o, value)


def uint64_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_uint64(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_UINT64, value)
    set_data(o, value)


def int64_to_data(o, offset, transform, tag_id):
    value = plc_tag_get_int64(tag_id, offset)
    if value is None:
        raise ValueError
    if transform is not None:
        value = run_transform(transform, DATA_TYPE_INT64, value)
    set_data(o, value)


def init(cfg_proto, cfg_pull, timeout=5):
    global plc_timeout, plc_path

    jsonschema.validate(cfg_proto, SCHEMA_PROTO)
    jsonschema.validate(cfg_pull, SCHEMA_PULL)
    if cfg_proto['name'] in ['enip/ab_eip']:
        try:
            host, port = cfg_proto['source'].rsplit(':', 1)
        except:
            host = cfg_proto['source']
            port = 44818
    else:
        raise ValueError(f'Unsupported protocol: {cfg_proto["name"]}')

    path = cfg_proto.get('path', '')

    cpu = cfg_proto['cpu']

    plc_timeout = timeout * 1000

    plc_path = f'protocol=ab_eip&gateway={host}:{port}&path={path}&cpu={cpu}'

    for p in cfg_pull:
        tag = p['tag']
        size = p.get('size', 1)
        count = p.get('count')
        tag_path = f'&elem_size={size}'
        if count:
            tag_path += f'&elem_count={count}'
        tag_path += f'&name={tag}'
        pmap = []
        for m in p.get('process', []):
            offset = m['offset']
            o = m['set-id']
            tp = m.get('type')
            transform = m.get('transform')
            offset = parse_int(offset)
            if tp in ['real', 'real32']:
                fn = partial(real32_to_data, o, offset,
                             prepare_transform(o, transform))
            elif tp in ['real64']:
                fn = partial(real64_to_data, o, offset,
                             prepare_transform(o, transform))
            elif tp in ['uint8', 'byte']:
                fn = partial(uint8_to_data, o, offset,
                             prepare_transform(o, transform))
            elif tp in ['uint16', 'word']:
                fn = partial(uint16_to_data, o, offset,
                             prepare_transform(o, transform))
            elif tp in ['uint32', 'dword']:
                fn = partial(uint32_to_data, o, offset,
                             prepare_transform(o, transform))
            elif tp in ['uint64', 'qword']:
                fn = partial(uint64_to_data, o, offset,
                             prepare_transform(o, transform))
            elif tp in ['sint8', 'int8']:
                fn = partial(int8_to_data, o, offset,
                             prepare_transform(o, transform))
            elif tp in ['sint16', 'int16']:
                fn = partial(int16_to_data, o, offset,
                             prepare_transform(o, transform))
            elif tp in ['sint32', 'int32']:
                fn = partial(int32_to_data, o, offset,
                             prepare_transform(o, transform))
            elif tp in ['sint64', 'int64']:
                fn = partial(int64_to_data, o, offset,
                             prepare_transform(o, transform))
            else:
                raise ValueError(f'type unsupported: {tp}')
            pmap.append(fn)
        register_puller(partial(read_tag, tag_path), pmap)


def shutdown():
    for p, t in active_tags.items():
        plc_tag_destroy(t)
    active_tags.clear()



# TODO tests

__author__ = 'Altertech, https://www.altertech.com/'
__copyright__ = 'Copyright (C) 2020 Altertech'
__license__ = 'Apache License 2.0'

__version__ = '0.1.2'

import sys
import argparse
import importlib
import threading
import jsonschema

import yaml

from time import perf_counter, sleep
from .outputs import OUTPUT_METHODS, eprint, print_trace, set_time_format
from queue import Queue

q = Queue()

processor = None

data = {}

pullers = []
processor_maps = []

DEFAULT_TIMEOUT = 5
DEFAULT_FREQUENCY = 1
DEFAULT_BEACON_INTERVAL = 0

last_pull_time = 0

config = {
    'timeout': DEFAULT_TIMEOUT,
    'freq': DEFAULT_FREQUENCY,
    'beacon': DEFAULT_BEACON_INTERVAL,
    'output': None
}

CONFIG_SCHEMA = {
    'type': 'object',
    'properties': {
        'version': {
            'type': 'integer',
            'minimum': 1
        },
        'timeout': {
            'type': 'number',
            'minimum': 0
        },
        'beacon': {
            'type': 'number',
            'minimum': 0
        },
        'freq': {
            'type': 'number',
            'minimum': 0
        },
        'proto': {
            'type': 'object'
        },
        'output': {
            'type': ['string', 'null']
        },
        'time-format': {
            'type': 'string',
            'enum': ['iso', 'timestamp']
        },
        'pull': {
            'type': 'array'
        }
    },
    'additionalProperties': False,
    'required': ['version', 'proto', 'pull']
}

output = None


def register_puller(fn, pmap=[]):
    pullers.append((fn, pmap))


def get_last_pull_time():
    return last_pull_time


def clear():
    data.clear()
    pullers.clear()
    processor_maps.clear()
    from .dp import clear
    dp.clear()


def set_data(o, value):
    if value is None:
        return
    current = data.get(o)
    if current != value:
        data[o] = value
        output(o, value)


def _t_beacon(fn, interval):
    try:
        next_beacon = perf_counter() + interval
        while True:
            ts = next_beacon - perf_counter()
            if ts > 0:
                sleep(ts)
            fn()
            next_beacon += interval
    except:
        print_trace()


def _t_processor():
    try:
        while True:
            try:
                data, prc_map = q.get()
            except TypeError:
                break
            for fn in prc_map:
                fn(data)
    except:
        print_trace()


def do(loop=False):

    interval = config['interval']

    next_iter = perf_counter() + interval

    def pull_and_process():
        nonlocal next_iter
        global last_pull_time

        last_pull_time = perf_counter()

        for plr, prc_map in pullers:
            q.put((plr(), prc_map))
        if not processor.is_alive():
            raise RuntimeError('processor thread is gone')
        ts = next_iter - perf_counter()
        if loop:
            if ts > 0:
                sleep(ts)
            else:
                eprint('WARNING: main loop timeout')
        next_iter += interval

    if loop:
        while True:
            pull_and_process()
    else:
        pull_and_process()


def main():
    global output
    global processor
    ap = argparse.ArgumentParser()
    ap.add_argument('-F',
                    '--config',
                    help='Configuration file',
                    metavar='CONFIG',
                    required=True)
    ap.add_argument('-L',
                    '--loop',
                    help='Loop (production)',
                    action='store_true')
    ap.add_argument('-R',
                    '--auto-restart',
                    help='Auto restart loop on errors',
                    action='store_true')
    a = ap.parse_args()

    with open(a.config) as fh:
        config.update(yaml.safe_load(fh))

    jsonschema.validate(config, CONFIG_SCHEMA)

    config['interval'] = 1 / config['freq']

    set_time_format(config.get('time-format'))

    try:
        om = OUTPUT_METHODS[config['output']]
    except KeyError:
        raise Exception('Unsupported output type')
    output = om['output']
    send_beacon = om.get('beacon')

    proto = config['proto']['name']

    if '/' in proto:
        proto = proto.split('/', 1)[0]

    lib = importlib.import_module(f'pulr.proto.{proto}')

    if a.loop and send_beacon and config['beacon']:
        threading.Thread(target=_t_beacon,
                         name='beacon',
                         args=(send_beacon, config['beacon']),
                         daemon=True).start()

    while True:
        clear()
        try:
            lib.init(config['proto'],
                     config.get('pull', []),
                     timeout=config['timeout'])

            processor = threading.Thread(target=_t_processor,
                                         name='processor',
                                         daemon=True)
            processor.start()
            try:
                do(loop=a.loop)
            finally:
                # finish data processing before shutting down proto lib
                q.put(None)
                processor.join()
                lib.shutdown()
            if not a.loop:
                break
        except KeyboardInterrupt:
            break
        except:
            if a.auto_restart and a.loop:
                print_trace()
                sleep(config['interval'])
            else:
                raise
